/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

#ifndef FGGL_GFX_PAINT_HPP
#define FGGL_GFX_PAINT_HPP

#include <string>
#include <vector>
#include <fggl/math/types.hpp>

namespace fggl::gfx {

	using RadianAngle = float;

	namespace colours {
		constexpr math::vec3 WHITE{1.0F, 1.0F, 1.0F};
		constexpr math::vec3 BLACK{0.0F, 0.0F, 0.0F};

		constexpr math::vec3 RED{1.0F, 0.0F, 0.0F};
		constexpr math::vec3 GREEN{0.0F, 1.0F, 0.0F};
		constexpr math::vec3 BLUE{0.0F, 0.0F, 1.0F};

		constexpr math::vec3 YELLOW{1.0F, 1.0F, 0.0F};
		constexpr math::vec3 CYAN{0.0F, 1.0F, 1.0F};
		constexpr math::vec3 MAGENTA{1.0F, 0.0F, 1.0F};
	}


	enum class PathType {
		MOVE,
		PATH,
		BAZIER2,
		BAZIER3,
		COLOUR,
		CLOSE
	};

	struct Path2D {
		inline explicit Path2D(math::vec2 start) : m_points(), m_types() {
			moveTo(start);
		}

		inline void moveTo(math::vec2 pos) {
			m_points.push_back(pos);
			m_types.push_back(PathType::MOVE);
		}

		inline void pathTo(math::vec2 pos) {
			m_points.push_back(pos);
			m_types.push_back(PathType::PATH);
		}

		void bezierTo(math::vec2 cp1, math::vec2 pos);
		void bezierTo(math::vec2 cp1, math::vec2 cp2, math::vec2 pos);

		void arc(math::vec2 center, float radius, RadianAngle startAngle, RadianAngle endAngle, bool ccw);

		void colour(math::vec3 colour) {
			m_colours.push_back(colour);
			m_types.push_back(PathType::COLOUR);
		}

		void close() {
			pathTo(m_points[0]);
			m_types.push_back(PathType::CLOSE);
		}

		std::vector<math::vec2> m_points;
		std::vector<PathType> m_types;
		std::vector<math::vec3> m_colours;
	};

	inline Path2D make_rect(math::vec2 center, math::vec2 radius, math::vec3 colour = {1.0f, 1.0f, 1.0f}) {
		fggl::gfx::Path2D tileGfx(center);
		tileGfx.colour(colour);
		tileGfx.moveTo(center - radius);
		tileGfx.pathTo({center.x + radius.x, center.y - radius.y});
		tileGfx.pathTo(center + radius);
		tileGfx.pathTo({center.x - radius.x, center.y + radius.y});
		tileGfx.close();
		return tileGfx;
	}

	struct ShapeOpts {
		float angleOffset = 0.0F;
		bool sinFirst = true;
	};

	inline Path2D make_shape(math::vec2 center, float radius, int sides, math::vec3 colour = {1.0f, 1.0f, 1.0f}, ShapeOpts opts = {}) {
		double angle = (M_PI * 2.0) / sides;

		fggl::gfx::Path2D tileGfx(center);
		tileGfx.colour(colour);

		for (int i=0; i < sides; ++i) {
			float xPos = (float)(sin(i * angle + opts.angleOffset) * radius) + center.x;
			float yPos = (float)(cos(i * angle + opts.angleOffset) * radius) + center.y;
			if (!opts.sinFirst) {
				std::swap(xPos, yPos);
			}
			if ( i == 0 ) {
				tileGfx.moveTo( {xPos, yPos} );
			} else {
				tileGfx.pathTo({xPos, yPos});
			}
		}
		tileGfx.close();
		return tileGfx;
	}

	enum class PaintType {
		FILL,
		STROKE
	};

	struct PaintCmd {
		PaintType type;
		Path2D path;
	};

	struct TextCmd {
		const std::string text;
		const math::vec2 pos;
	};

	class Paint {
		public:
			Paint() = default;
			Paint(Paint &paint) = delete;

			inline void fill(Path2D &path) {
				m_cmds.push_back({PaintType::FILL, path});
			}

			void stroke(Path2D &path) {
				m_cmds.push_back({PaintType::STROKE, path});
			}

			void text(const std::string &text, const math::vec2 &pos) {
				m_text.push_back({text, pos});
			}

			const std::vector<PaintCmd> &cmds() const {
				return m_cmds;
			}

			const std::vector<TextCmd> &textCmds() const {
				return m_text;
			}

		private:
			std::vector<PaintCmd> m_cmds;
			std::vector<TextCmd> m_text;
	};

}

#endif
